from empath import Empath
from twitter_demographer.components import Component


class EmpathAnalyzer(Component):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def outputs(self):
        return [f"empath_{k}" for k in Empath().analyze("test").keys()]

    def inputs(self):
        return ["text"]

    def infer(self, data):
        lexicon = Empath()
        results = self.initialize_return_dict()

        for tweet in data["text"].values.tolist():
            counts = lexicon.analyze(tweet, normalize=True)

            for item, value in counts.items():
                results[f"empath_{item}"].append(value)

        return results



from transformers import Trainer
from twitter_demographer.support.dataset import prepare_dataset
from transformers import AutoTokenizer, AutoModelForSequenceClassification
import pandas as pd
from datasets import Dataset
import numpy as np
from twitter_demographer.components import Component
from twitter_demographer.components import not_null
import transformers
import datasets
import logging

datasets.logging.disable_progress_bar()
transformers.logging.disable_progress_bar()

class HuggingFaceClassifier(Component):
    """
    Wraps a huggingface ForSequenceClassification model
    """

    def __init__(self, model_name, logger_level=logging.CRITICAL):
        self.model_name = model_name
        super().__init__()
        requests_logger = logging.getLogger('transformers')
        requests_logger.setLevel(logger_level)

    def inputs(self):
        return ["text"]

    def outputs(self):
        return [f"{self.model_name}"]

    @not_null("text")
    def infer(self, data):

        tokenizer = AutoTokenizer.from_pretrained(self.model_name)
        model = AutoModelForSequenceClassification.from_pretrained(self.model_name)

        df = pd.DataFrame({"texts": data["text"].values.tolist()})

        train_dataset = Dataset.from_pandas(df)
        train_dataset = prepare_dataset(train_dataset, tokenizer)

        transformers.logging.set_verbosity_error()

        trainer = Trainer(model=model)

        local_results = np.argmax(trainer.predict(train_dataset)[0], axis=1)

        return {self.model_name: local_results}



import hashlib
from tqdm.auto import tqdm
import datetime
import json
import random



class Demographer:
    """
    Main entry point of twitter-demographer.
    """

    def __init__(self, cache_steps=False, label_swapping=True):
        self.components = []
        self.cache_steps = cache_steps
        self.local_inputs = []
        self.label_swapping = label_swapping

    def add_component(self, component):
        if len(self.components) == 0:
            self.components.append(component)
            self.local_inputs.extend(component.outputs() + component.inputs())  # first run we add both
        else:
            if set(component.inputs()) <= set(self.local_inputs):
                self.components.append(component)
            else:
                raise Exception("Some of the fields are not supported by this composition")

    def infer(self, data):
        pbar = tqdm(total=len(self.components), position=0)
        data = data.copy()

        for index, component in enumerate(self.components):
            pbar.set_description(f"Running Demographer: Component {index + 1}")
            update = component.infer(data)
            for key, value in update.items():
                data[key] = value

            if self.cache_steps:
                data.to_parquet("local.csv", engine="pyarrow")

            del component
            pbar.update(1)

        # add hash(y: str) because hash(x: int) = x
        data["screen_name"] = data["screen_name"].apply(
            lambda x: hashlib.sha3_256(str(hash(x) + hash("hash")).encode()).hexdigest())
        pbar.close()
        drop_keys = ['tweet_ids', 'name', 'user_id_str',
                     'user_id', 'id', 'profile_image_url', 'description']

        for key in drop_keys:
            del data[key]

        data = data.sample(frac=1)

        if len(data) > 100:
            if self.label_swapping:
                noisy = int((len(data) * 0.01)/2) + 1
                for _ in range(0, noisy):
                    col = random.choice(data.columns)
                    idx1, idx2 = random.sample(data.index.tolist(), 2)

                    data[col].iloc[idx1] = data[col].iloc[idx2]
                return data
            else:
                return data

        return data

    def get_versioned_json(self):
        json_dict = {"run_at": datetime.datetime.now().time(), "components" : []}

        for component in self.components:
            json_dict["components"].append({type(component).__name__ : component.__dict__})

        return json.dumps(json_dict)



import numpy as np
import pandas as pd
from contextualized_topic_models.models.ctm import ZeroShotTM
from contextualized_topic_models.utils.data_preparation import TopicModelDataPreparation
from contextualized_topic_models.utils.preprocessing import WhiteSpacePreprocessingStopwords
import re
from twitter_demographer.components import Component, not_null
from twitter_demographer.support.utils import twitter_preprocess

class TopicModeler(Component):
    """
    A simple topic modeler based on Contextualized Topic Models. This topic model is multilingual by default

    see: https://aclanthology.org/2021.acl-short.96/
    """

    def outputs(self):
        return ["topic", "topic_words"]

    def inputs(self):
        return ["text"]

    def __init__(self, topic_num, embedding_model="paraphrase-multilingual-mpnet-base-v2", stopwords=None):
        super().__init__()
        self.topic_num = topic_num
        self.embedding_model = embedding_model
        self.stopwords = stopwords if stopwords else []

    @not_null("text")
    def infer(self, data):
        df = pd.DataFrame({"text": data["text"].values.tolist()})
        df["text"] = df["text"].apply(twitter_preprocess)
        df["text"] = df["text"].apply(lambda x : re.sub(r'\s*(?:https?://)?www\.\S*\.[A-Za-z]{2,5}\s*', '', x))

        documents = df["text"].values.tolist()

        sp = WhiteSpacePreprocessingStopwords(documents, stopwords_list=self.stopwords)
        preprocessed_documents, unpreprocessed_corpus, vocab = sp.preprocess()

        tp = TopicModelDataPreparation(self.embedding_model)

        training_dataset = tp.fit(text_for_contextual=unpreprocessed_corpus, text_for_bow=preprocessed_documents)

        ctm = ZeroShotTM(bow_size=len(tp.vocab), contextual_size=768, n_components=self.topic_num, num_epochs=20)
        ctm.fit(training_dataset)
        topics_predictions = np.argmax(ctm.get_thetas(training_dataset, n_samples=15), axis=1).tolist()
        topic_list = ctm.get_topic_lists(5)

        return {"topic": topics_predictions, "topic_words" : [topic_list[k] for k in topics_predictions]}



import abc
import numpy as np
import pandas as pd
import random
from tqdm.auto import tqdm
import tweepy
from tweepy.errors import BadRequest
import transformers
from abc import ABC
from twitter_demographer.support.utils import chunks

transformers.logging.set_verbosity(transformers.logging.ERROR)

class Component(ABC):
    """
    Abstract component class
    """

    def __init__(self, **kwargs):
        self.output_parameters = self.outputs()

    @abc.abstractmethod
    def outputs(self):
        pass

    @abc.abstractmethod
    def inputs(self):
        pass

    @abc.abstractmethod
    def infer(self, *args):
        pass

    def initialize_return_dict(self):
        """
        A simple function that creates a dictionary with keys = parameters and values empty lists
        :return:
        """
        results = {}

        for param in self.output_parameters:
            results[param] = []
        return results


class Rehydrate(Component):

    def __init__(self, bearer_token):
        """
        :param bearer_token: token from the twitter developer page
        """
        super().__init__()
        self.api = tweepy.Client(bearer_token, wait_on_rate_limit=True)

    def inputs(self):
        return ["tweet_ids"]

    def outputs(self):
        return ['screen_name', 'name', 'location', 'user_id_str',
                'user_id', 'id', 'profile_image_url', 'description',
                'created_at', 'text']

    def infer(self, data):


        tweet_ids = data["tweet_ids"].values.tolist()
        pbar = tqdm(total=len(data), position=1)

        good_tweets = {}

        for split in chunks(tweet_ids, 99):
            split = ",".join(list(map(str,split)))

            requested = self.api.get_tweets(split,
                                           tweet_fields=[
                                               'id', 'text', 'author_id', 'created_at', 'conversation_id',
                                               'entities',
                                               'public_metrics', 'referenced_tweets'
                                           ],
                                           expansions=[
                                               'author_id', 'referenced_tweets.id',
                                               'referenced_tweets.id.author_id',
                                           ],
                                           place_fields=['full_name', 'id'],
                                           user_fields=[
                                               'id', 'name', 'username', 'profile_image_url', 'description',
                                               'location'
                                           ])

            user_mapping_dict = {k["id"]: k for k in requested[1]["users"]}
            tweets = requested[0]

            for tweet in tweets:
                results = {}

                pbar.set_description("Running Hydrate")
                user_id = tweet["author_id"]

                user_includes = user_mapping_dict[user_id]
                results["text"] = tweet.text
                results["description"] = user_includes.description
                results["created_at"] = tweet.created_at
                results["name"] = user_includes.name
                results["screen_name"] = user_includes.username
                results["user_id"] = user_includes.id
                results["id"] = user_includes.id
                results["user_id_str"] = str(user_includes.id)
                results["profile_image_url"] = user_includes.profile_image_url.replace("_normal", "")
                results["location"] = user_includes.location
                good_tweets[tweet.id] = results

            # we know get all the tweets that we were not able to reconstruct
            errors = requested[2]
            for error in errors:
                pbar.update(1)
                results = {}

                for field in self.outputs():
                    results[field] = (None)

                good_tweets[int(error["value"])] = results

            pbar.update(99)
        pbar.close()

        results = self.initialize_return_dict()
        for tw_id in data["tweet_ids"].values.tolist():
            for field in self.outputs():
                results[field].append(good_tweets[int(tw_id)][field])

        return results

def not_null(param):
    """
    k = pd.DataFrame({"text" : ["a", None, "b"]})

    @not_null("text")
    def method(data):
        return {"new_label" : [1 for _ in data.iterrows()]}

    print(class_.method(k)) # index at 1 should be None

    {'new_label': [1, None, 1]}

    """
    def Inner(func):
        def wrapper(*args, **kwargs):

            data = args[1]
            to_function = data[~(data[param].isna())]
            args = (args[0],) + (to_function,) + args[2:]
            resulted = func(*args, **kwargs)

            new_dict = dict()

            for keys in resulted:
                new_dict[keys] = [None] * len(data)
                for index, value in zip(to_function.index.values.tolist(), resulted[keys]):
                    new_dict[keys][index] = value

            return new_dict
        return wrapper
    return Inner

def repropagate_on_duplicates(param):
    """
    k = pd.DataFrame({"text" : ["a", None, "b"]})


    """
    raise Exception("Not Implemented")
    def Inner(func):
        def wrapper(*args, **kwargs):

            data = args[1]
            to_function = data[~(data[param].isna())]
            args = (args[0],) + (to_function,) + args[2:]
            resulted = func(*args, **kwargs)

            new_dict = dict()

            for keys in resulted:
                new_dict[keys] = [None] * len(data)
                for index, value in zip(to_function.index.values.tolist(), resulted[keys]):
                    new_dict[keys][index] = value

            return new_dict
        return wrapper
    return Inner

def label_swapper(dataframe, percentage=0.01):
    msk = np.random.rand(len(dataframe)) < percentage
    shuffle = dataframe[msk]
    rest = dataframe[~msk]

    col = random.choice(shuffle.cloumns)

    vals = shuffle[col].values.tolist()
    random.shuffle(vals)
    shuffle[col] = vals

    dataframe = pd.concat([rest, shuffle]).sample(frac=1)

    return dataframe





import time
from twitter_demographer.components import Component
import requests
from typing import Union
import json
from typing import List
from twitter_demographer.components import not_null
from tqdm import tqdm


class PerspectiveScorer(Component):
    """
    This component uses the perspective API to annotate tweets
    https://developers.perspectiveapi.com/s/about-the-api
    """

    def __init__(self, api_key: str, attributes: Union[List[str], None] = None, wait_seconds: int = 2, **kwargs):
        """

        :param api_key: API KEY provided by https://developers.perspectiveapi.com
        :param attributes: list of the attributes to use from the perspective API
        :param wait_seconds: how many seconds to wait for each request. This varies if you have the free quota or not
        :param kwargs:
        """
        self.api_key = api_key
        if attributes is None:
            self.attributes = "TOXICITY,SEVERE_TOXICITY,IDENTITY_ATTACK,INSULT,PROFANITY,THREAT".split(",")
        else:
            self.attributes = attributes

        self.wait_seconds = wait_seconds
        super().__init__(**kwargs)

    def outputs(self):
        return [f"perspective_{k}" for k in self.attributes]

    def inputs(self):
        return ["text"]

    @not_null("text")
    def infer(self, data, *args):

        results = self.initialize_return_dict()

        pbar = tqdm(total=len(data), position=1)
        pbar.set_description("Running Toxicity API")
        for text in data["text"].values.tolist():
            pbar.update(1)

            try:
                dici = self.score(text)

                for index, values in dici.items():
                    results[f"perspective_{index}"].append(dici[index])
            except:
                for val in self.outputs():
                    results[val].append(None)

            time.sleep(self.wait_seconds)
        pbar.close()
        return results

    def score(self, text):
        headers = {
            'Content-Type': 'application/json',
        }

        params = (
            ('key', self.api_key),
        )

        data = {"comment": {"text": text}, "requestedAttributes": {}}

        for att in self.attributes:
            data["requestedAttributes"][att] = {}

        data = json.dumps(data)
        response = requests.post('https://commentanalyzer.googleapis.com/v1alpha1/comments:analyze', headers=headers,
                                 params=params, data=data)

        return {index: response.json()["attributeScores"][index]["summaryScore"]["value"] for index in self.attributes}



"""
Code has been imported from the M3Inference Library.
"""
import urllib.request
from io import BytesIO
from PIL import Image
from appdirs import user_cache_dir
import os
from multiprocessing import Pool
import multiprocessing

def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]

def twitter_preprocess(text):
    """
    Stole this from: https://huggingface.co/cardiffnlp/twitter-roberta-base-offensive?text=I+think+they+are+a+bunch+of+losers.
    :param text:
    :return:
    """
    new_text = []
    for t in text.split(" "):
        t = '@user' if t.startswith('@') and len(t) > 1 else t
        new_text.append(t)
    return " ".join(new_text)

def get_extension(img_path):
    dotpos = img_path.rfind(".")
    extension = img_path[dotpos + 1:]
    if extension.lower() == "gif":
        return "png"
    return extension

def download_resize_img(url, img_out_path, img_out_path_fullsize=None):
    # url=url.replace("_200x200","_400x400")
    try:
        img_data = urllib.request.urlopen(url)
        img_data = img_data.read()

        if img_out_path_fullsize != None:
            with open(img_out_path_fullsize, "wb") as fh:
                fh.write(img_data)

    except:
        print("Something wrong happened in downloading the image")
        return False

    return resize_img(BytesIO(img_data), img_out_path, force=True)

def resize_img(img_path, img_out_path, filter=Image.BILINEAR, force=False):

    try:
        img = Image.open(img_path).convert("RGB")
        if img.size[0] + img.size[1] < 400 and not force:
            print(f'{img_path} is too small. Skip.')
            return
        img = img.resize((224, 224), filter)
        img.save(img_out_path)
        return True

    except Exception as e:
        print(f'Error when resizing {img_path}\nThe error message is {e} for {img_out_path}\n')
        return False



from twitter_demographer.components import Component, not_null
import re
import liwc
from collections import Counter

def liwc_tokenize(text):
    text = text.lower()
    for match in re.finditer(r'\w+', text, re.UNICODE):
        yield match.group(0)

class LIWCAnalyzer(Component):
    """
    A wrapper over LIWC files
    """

    def __init__(self, liwc_dic, **kwargs):
        """

        :param liwc_dic: this is the liwc file. The method exepcts something like "LIWC2007_English100131.dic"
        :param kwargs:
        """

        parse, category_names = liwc.load_token_parser(liwc_dic)
        self.category_names = category_names
        self.parse = parse
        super().__init__(**kwargs)

    def outputs(self):
        return [f"LIWC_{k}" for k in self.category_names]

    def inputs(self):
        return ["text"]

    @not_null("text")
    def infer(self, data, *args):

        results = self.initialize_return_dict()

        for text in data["text"].values.tolist():
            tokens = liwc_tokenize(text)

            counts = dict(Counter(category for token in tokens for category in self.parse(token)))

            for item, value in counts.items():
                results[f"LIWC_{item}"].append(value)

        return results




#!/usr/bin/env python

"""Tests for `twitter_demographer` package."""

import pytest
import os
from twitter_demographer import twitter_demographer
from twitter_demographer.geolocation.nominatim import NominatimDecoder
from twitter_demographer.components import Rehydrate
import pandas as pd
from twitter_demographer.analyzer.liwc import LIWCAnalyzer

def test_demographer():
    demo = twitter_demographer.Demographer()
    data = pd.DataFrame({"tweet_ids": ["1477976329710673921", "1467887350084689928"]})

    re = Rehydrate(os.getenv('TWEEPY_BEARER_TOKEN'))
    geo = NominatimDecoder()

    demo.add_component(re)
    demo.add_component(geo)

    demo.infer(data)

def test_liwc():
    """
    For this test we replicate the test run by the package liwc-python
    :return:
    """
    demo = twitter_demographer.Demographer()
    le = LIWCAnalyzer("tests/alpha.dic")

    data = pd.DataFrame(
        {"text": ["Any alpha bravo charlie Bravo"], "screen_name": ["a"], "tweet_ids": [1], "name": ["k"],
         "user_id_str": [12], "user_id": [12], "id": [12], "profile_image_url": ["hht"], "description": ["asdafs"]})

    demo.add_component(le)

    data = demo.infer(data)

    assert data["LIWC_A"].values[0] == 2
    assert data["LIWC_Bravo"].values[0] == 2





from tqdm.auto import tqdm
import geocoder
from twitter_demographer.components import Component
import logging
from twitter_demographer.components import not_null
import time

class NominatimDecoder(Component):
    """
    Wrappers on the geocoder API to disambiguate users' locations using nominatim from open street map
    """

    def __init__(self, server_url="https://nominatim.openstreetmap.org/search", sleep_time=1.5, logger_level=logging.CRITICAL):
        super().__init__()
        self.server_url = server_url
        self.sleep_time = sleep_time
        requests_logger = logging.getLogger('geocoder')
        requests_logger.setLevel(logger_level)

    def outputs(self):
        return ["nominatim_city", "nominatim_country"]

    def inputs(self):
        return ["location"]

    @not_null("location")
    def infer(self, data):

        geo = self.initialize_return_dict()
        pbar = tqdm(total=len(data), position=1)
        pbar.set_description("Geocoder")

        for val in data["location"]:

            if val is None:
                geo["nominatim_country"].append(None)
                geo["nominatim_city"].append(None)
            else:

                g = geocoder.osm(val, server_url=self.server_url).osm

                if g is not None and "addr:country" in g:
                    geo["nominatim_country"].append(g["addr:country"])
                else:
                    geo["nominatim_country"].append(None)

                if g is not None and "addr:city" in g:
                    geo["nominatim_city"].append(g["addr:city"])
                else:
                    geo["nominatim_city"].append(None)
                time.sleep(self.sleep_time)

            pbar.update(1)
        pbar.close()

        return geo
